from __future__ import annotations

import hashlib
import typing as t

import numpy as np

from nnviz import colors


class _BubbleTree:
    def __init__(
        self,
        origin: np.ndarray,
        radius: float,
        children: t.Dict[t.Hashable, _BubbleTree],
        factor: float = 0.3,
        ord: int = 2,
        budget: int = 100,
        rate: float = 0.01,
    ) -> None:
        self.origin = origin.astype(np.float64)
        self.radius = float(radius)
        self.children = children

        self._factor = factor
        self._ord = ord
        self._budget = budget
        self._rate = rate

    def __getitem__(self, key: t.Hashable) -> _BubbleTree:
        return self.children[key]

    def __contains__(self, key: t.Hashable) -> bool:
        return key in self.children

    def _pseudo_rand(self, arr: np.ndarray, n: int) -> np.ndarray:
        # Pick a pseudo-random direction based on the hash of the children_origins
        hashable = arr.tobytes()
        hash_ = hashlib.sha256(hashable).hexdigest()

        # Repeat the hash until we have enough bits to fill the direction
        n_bytes = 8
        while len(hash_) < arr.shape[1] * n_bytes:  # pragma: no cover
            hash_ += hashlib.sha256(hash_.encode("utf-8")).hexdigest()

        # Convert the hash to a valid direction
        prand = np.array(
            [
                np.float64.fromhex(hash_[i : i + n_bytes])
                for i in range(0, len(hash_), n_bytes)
            ]
        )
        return np.sin(prand[:n])

    def _starting_point_proposal(
        self, children_origins: np.ndarray, children_radii: np.ndarray
    ) -> np.ndarray:
        # Compute the pairwise distances between the children
        distances = np.linalg.norm(
            children_origins[:, None, :] - children_origins[None, :, :],
            ord=self._ord,
            axis=2,
        )
        distances -= children_radii[:, None] + children_radii[None, :]

        # Select the child with the largest distance to the closest child
        starting_child_index = np.argmax(
            np.min(distances + np.eye(distances.shape[0]) * self.radius, axis=1)
        )
        start_child_org = children_origins[starting_child_index]
        start_child_rad = children_radii[starting_child_index]

        # Compute a pseudo-random direction
        direction = self._pseudo_rand(children_origins, children_origins.shape[1])

        # Normalize the direction so that it has length children radius
        direction = (
            direction / np.linalg.norm(direction, ord=self._ord) * start_child_rad
        )

        # Place the new subtree at the origin of the random child plus the direction
        return start_child_org + direction

    def spawn(self, key: t.Hashable) -> None:
        # If there are no children, we can just return a new subtree with the origin
        # as the origin of the new subtree but smaller radius
        if len(self.children) == 0:
            spawned = _BubbleTree(self.origin, self.radius * self._factor, {})
            self.children[key] = spawned
            return

        # Init the arrays
        children_origins = np.stack([x.origin for x in self.children.values()])
        children_radii = np.stack([x.radius for x in self.children.values()])
        eps = self.radius * 1e-5

        # Find a good starting point
        new_origin = self._starting_point_proposal(children_origins, children_radii)

        budget = self._budget
        while budget > 0:
            # Compute the distance to the edge of the parent tree
            distance_to_edge = self.radius - np.linalg.norm(
                new_origin - self.origin, ord=self._ord  # type: ignore
            )

            # Compute the distance to the closest child
            distances = np.linalg.norm(
                new_origin - children_origins, ord=self._ord, axis=1  # type: ignore
            )

            # We have to subtract the radius of the child from the distance to the child
            # because the distance is computed from the center of the child
            distances -= children_radii

            # Select the closest child
            min_index, min_distance = np.argmin(distances), np.min(distances)

            # If the distance to the edge is smaller than the half distance to the closest
            # child, the direction is towards the center of the parent tree
            if distance_to_edge < min_distance:
                direction = self.origin - new_origin

            # If the distance to the edge is larger than the distance to the closest
            # child, the direction is away from the center of the closest child
            else:
                direction = new_origin - children_origins[min_index]

            # Normalize the direction
            direction /= np.linalg.norm(direction, ord=self._ord) + eps  # type: ignore

            # Move the new origin in the direction
            new_origin += direction * self.radius * self._rate

            # Reduce the budget
            budget -= 1

        # The new radius is the distance to the closest child
        new_radius = min(min_distance, distance_to_edge)  # type: ignore

        # Clamp the raidus to [eps, self.radius * self._factor]
        new_radius = max(eps, min(new_radius, self.radius * self._factor))

        spawned = _BubbleTree(new_origin, new_radius, {})
        self.children[key] = spawned


class BubbleColorPicker(colors.ColorPicker):
    """A color picker that places the colors in a tree structure and picks colors
    based on the path to the color.
    """

    def __init__(self) -> None:
        super().__init__()

        self._color_tree = _BubbleTree(np.array([127, 127, 127]), 127, {})
        self._color_tree.spawn("__root__")
        self._ord = "fro"

    def _np_to_rgb(self, np_color: np.ndarray) -> colors.RGBColor:
        return colors.RGBColor(*np_color.astype(np.uint8).tolist())

    def _pick_recur(self, sub_tree: _BubbleTree, *args: t.Hashable) -> colors.RGBColor:
        # If there are no arguments left, we have reached the end of the path and we can
        # return the color at the end of the path
        if len(args) == 0:
            return self._np_to_rgb(sub_tree.origin)

        # If the first argument is not in the sub-tree, we need to select a new color
        # close to the origin color and add it to the sub-tree
        if args[0] not in sub_tree:
            sub_tree.spawn(args[0])

        # Recurse on the sub-tree with the first argument removed
        return self._pick_recur(sub_tree[args[0]], *args[1:])

    def pick(self, *args: t.Hashable) -> colors.RGBColor:
        return self._pick_recur(self._color_tree, *args)


if __name__ == "__main__":  # pragma: no cover
    import random

    import cv2 as cv

    cv.namedWindow("sasso", cv.WINDOW_NORMAL)
    H = W = 512
    cv.resizeWindow("sasso", H, W)

    bubble = _BubbleTree(np.array([0, 0]), 1.0, {})

    depth = 3
    possible_keys = [chr(i) for i in range(ord("a"), ord("a") + 21)]

    while True:

        canvas = np.zeros((H, W, 3))

        # Sample with replacement from the possible keys
        args = [random.choice(possible_keys) for _ in range(depth)]
        # Add some other random arguments
        args += [random.random() for _ in range(2)]

        def add_args(tree: _BubbleTree, args: t.List[t.Hashable]) -> None:
            if len(args) == 0:
                return

            if args[0] not in tree:
                tree.spawn(args[0])

            add_args(tree[args[0]], args[1:])

        add_args(bubble, args)  # type: ignore

        def visit(tree: _BubbleTree, d: int) -> None:
            row, col = int(tree.origin[0] * H // 2 + H // 2), int(
                tree.origin[1] * H // 2 + H // 2
            )
            if len(tree.children) == 0:
                canvas[row, col, :] = 255

            if d <= depth and d > 0:
                # Draw a circle around the origin of the tree
                cv.circle(
                    canvas,
                    (col, row),
                    int(tree.radius * H // 2),
                    (0, 0, 255 // d),
                    thickness=1,
                )

            for child in tree.children.values():
                visit(child, d + 1)

        visit(bubble, 0)

        canvas = (canvas / np.max(canvas) * 255).astype(np.uint8)
        cv.imshow("sasso", cv.cvtColor(canvas, cv.COLOR_RGB2BGR))
        key = cv.waitKey(1)

        if key == ord("q"):
            break



import importlib
import re
import sys
import typing as t
from pathlib import Path

import torch
import torch.nn as nn
import torchvision

from nnviz.inspection.base import NNInspector
from nnviz.inspection.torchfx import TorchFxInspector


def load_from_torchvision(model: str) -> nn.Module:
    return getattr(torchvision.models, model, None)()  # type: ignore


def load_from_file_import(src: str, symbol_name: str) -> t.Any:
    """Import a symbol from a python file. The file can be everywhere in the file system,
    it will be added to the python path.

    Args:
        src (str): The path to the python file.
        symbol_name (str): The name of the symbol to import.

    Returns:
        t.Any: The imported symbol.
    """
    path = Path(src)
    sys.path.append(str(path.parent))
    module = importlib.import_module(path.stem)
    return getattr(module, symbol_name)


def load_from_module_import(src: str, symbol_name: str) -> t.Any:
    """Import a symbol from a python module. The module must be importable from the
    current python path. The module can be a package.

    Args:
        src (str): The name of the module to import.
        symbol_name (str): The name of the symbol to import.

    Returns:
        t.Any: The imported symbol.
    """
    module = importlib.import_module(src)
    return getattr(module, symbol_name)


def load_from_dynamic_import(model: str) -> t.Any:
    """Import a symbol from a python file or module. The file or module can be anywhere
    in the file system, it will be added to the python path.

    Args:
        model (str): The path to the python file or module, followed by a colon and the

    Returns:
        t.Any: The imported symbol.
    """
    path, _, symbol_name = model.rpartition(":")
    if path.endswith(".py"):
        return load_from_file_import(path, symbol_name)
    else:
        return load_from_module_import(path, symbol_name)


def load_from_string(model: str) -> nn.Module:
    """Load a model from a string. The string can be either the name of a model in

    Args:
        model (str): The name of the model to load.

    Raises:
        ValueError: If the model could not be loaded.

    Returns:
        nn.Module: The loaded model.
    """
    if ":" not in model:
        smb = load_from_torchvision(model)
    else:
        smb = load_from_dynamic_import(model)

    if isinstance(smb, nn.Module):
        return smb

    if callable(smb):
        return smb()

    raise ValueError(f"Could not load model {model}")


DEFAULT_KEY = "x"
"""The default key to use when the input is a single tensor."""


def parse_input_str(in_str: t.Optional[str]) -> t.Optional[t.Dict]:
    """Parse an input string to a dictionary of tensors.

    Args:
        in_str (t.Optional[str]): The input string, can be:

            - None -> None
            - default -> float32 BCHW tensor of shape (1, 3, 224, 224) (commonly used)
            - image<side> (e.g. image224, image256, ...) -> float32 BCHH tensor
            - image<height>x<width> (e.g. image224x224, image256x512, ...) -> float32 BCHW tensor
            - tensor<s0>x<s1>x<s2>x... (e.g. tensor1x3x224x224, tensor1x3x256x512, ...) ->
            float32 generic tensors
            - <key1>:<value1>;<key2>:<value2>;... (e.g. x:tensor1x3x224x224;y:tensor1x3x256x512,
            ...) -> dictionary of tensors

    Returns:
        t.Optional[t.Dict]: A dictionary of tensors to use as input for the model.
    """
    # If the input is None, exit early returning None
    if in_str is None:
        return None

    def _r(istr: str) -> t.Any:
        # Presets syntax
        stx = {
            r"image(\d+)x(\d+)": lambda h, w: torch.rand(1, 3, int(h), int(w)),
            r"image(\d+)": lambda size: torch.rand(1, 3, int(size), int(size)),
            r"tensor(\d+(?:x\d+)*)": lambda shape: torch.rand(
                *[int(s) for s in shape.split("x")]
            ),
            r"((?:\w+):(?:.+);?)+": lambda etr: {
                k: _r(v) for k, v in re.findall(r"(\w+):([^;]+)", etr)
            },
            r"default": lambda: _r("image224"),
        }

        # if in_str matches one of the presets, return the result of the preset passing the
        # variable part of the string to the preset function
        for k, v in stx.items():
            match = re.match(k, istr)
            if match:
                return v(*match.groups())

        # If all else fails -> evil eval
        return eval(istr)

    # Parse the input string
    result = _r(in_str)

    # If the result is not a dict, wrap it in a dict with key "x"
    if not isinstance(result, t.Dict):
        result = {"x": result}

    return result



import typing as t
from pathlib import Path

import pygraphviz as pgv

import nnviz.drawing as drawing
from nnviz import colors, dataspec
from nnviz import entities as ent


class HTMLSpecVisitor(dataspec.DataSpecVisitor):
    """Visitor for the `DataSpec` class that produces an HTML table nesting."""

    def __init__(self) -> None:
        super().__init__()
        self._code = ""
        self._name_stack: t.List[t.List[str]] = []
        self._code_stack: t.List[str] = []

    def _end_visit(self, code: str) -> None:
        if len(self._name_stack) > 0:
            names = self._name_stack[-1]
            name = names.pop(0)  # <--- Pop from the START
            self._code += self._line([name, code])
            if len(names) == 0:
                self._name_stack.pop()
                self._end_composite_visit()
        else:
            self._code += code

    def _table_header(self) -> str:
        return '<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="0" WIDTH="220%">'

    def _cell(self, text: str) -> str:
        return f"<TD ALIGN='LEFT'>{text}</TD>"

    def _line(self, cells: t.Sequence[str]) -> str:
        return f"<TR>{''.join([self._cell(cell) for cell in cells])}</TR>"

    def _table(self, lines: t.Sequence[t.Sequence[str]]) -> str:
        return f'{self._table_header()}{"".join([self._line(line) for line in lines])}</TABLE>'

    def _begin_composite_visit(self, names: t.List[str]) -> None:
        self._name_stack.append(names)
        self._code_stack.append(self._code)
        self._code = self._table_header()

    def _end_composite_visit(self) -> None:
        code = self._code + "</TABLE>"
        self._code = self._code_stack.pop()
        self._end_visit(code)

    def visit_tensor_spec(self, spec: dataspec.TensorSpec) -> None:
        table = self._table(
            [["<B>shape:</B>", str(spec.shape)], ["<B>dtype:</B>", str(spec.dtype)]]
        )
        self._end_visit(table)

    def visit_builtin_spec(self, spec: dataspec.BuiltInSpec) -> None:
        table = self._table([["<B>type:</B>", spec.name]])
        self._end_visit(table)

    def visit_list_spec(self, spec: dataspec.ListSpec) -> None:
        self._begin_composite_visit([str(i) for i in range(len(spec.elements))])

    def visit_map_spec(self, spec: dataspec.MapSpec) -> None:
        self._begin_composite_visit(list(spec.elements.keys()))

    def visit_unknown_spec(self, spec: dataspec.UnknownSpec) -> None:
        self._code += self._table([["Unknown"]])

    @property
    def html(self) -> str:
        """Returns the HTML table."""
        return f"<{self._code}>"


def _human_format(num: t.Union[int, float]) -> str:
    """Returns a human readable format of a number."""
    magnitude = 0
    while abs(num) >= 1000:
        magnitude += 1
        num /= 1000.0
    # add more suffixes if you need them
    if magnitude == 0 and isinstance(num, int):
        return str(num)
    return "%.2f%s" % (num, ["", "K", "M", "G", "T", "P"][magnitude])


class GraphvizDrawer(drawing.GraphDrawer):
    """A graph drawer that uses Graphviz to draw a neural network graph."""

    def __init__(self, path: Path) -> None:
        # TODO: Remove this hard-code rodeo
        self._default_node_params = {
            "fontname": "Arial",
            "shape": "box",
            "style": "rounded,filled",
            "margin": "0.2,0.1",
            "color": "gray",
            "fillcolor": "gray",
        }
        self._default_edge_params = {
            "fontname": "Arial",
            "color": "black",
            "penwidth": "2.0",
        }
        self._default_subgraph_params = {
            "fontname": "Arial",
            "fontsize": "12",
            "style": "rounded,filled",
        }
        self._node_title_size = 24
        self._cluster_title_size = 18
        self._path = path
        # self._color_picker = colors.HashColorPicker()
        self._color_picker = colors.BubbleColorPicker()

        self._ignore_prefixes = [
            "torch.nn.",
        ]

    def _text_color(self, color: colors.RGBColor) -> str:
        return "black" if color.is_bright() else "white"

    def _multi_line(self, *lines: str) -> str:
        multi_line = "<BR/>".join(lines)
        return f"<{multi_line}>"

    def _pick_color_for_op_node(self, node: ent.OpNodeModel) -> colors.RGBColor:
        filtered_op = node.full_op
        # Check if there is a prefix to ignore
        for prefix in self._ignore_prefixes:
            if filtered_op.startswith(prefix):
                filtered_op = filtered_op[len(prefix) :]
                break
        pick_args = filtered_op.split(".") if filtered_op else []

        return self._color_picker.pick(*pick_args)

    def _op_params(self, node: ent.OpNodeModel) -> t.Dict[str, t.Any]:
        rgb = self._pick_color_for_op_node(node)
        color = rgb.to_hex()
        font_c = self._text_color(rgb)

        lines = [
            f'<B><FONT POINT-SIZE="{self._node_title_size}">{node.op}</FONT></B>',
            f"<I>{node.name}</I>",
        ]

        if node.n_parameters is not None and node.n_parameters > 0:
            lines.append("")
            lines.append(
                f"<B>params</B>: {_human_format(node.n_parameters)} ({_human_format(node.perc_parameters*100)}%)"  # type: ignore
            )

        if len(node.const_args) > 0:
            lines.append(
                "<B>args</B>: " + ", ".join([f"{arg}" for arg in node.const_args])
            )

        if len(node.const_kwargs) > 0:
            kwargs_line = "<B>kwargs</B>: " + ", ".join(
                [f"{key}={value}" for key, value in node.const_kwargs.items()]
            )
            lines.append(kwargs_line)

        lines.append("")
        lines.append(f"<B>src</B>: {node.full_op}")

        label = self._multi_line(*lines)
        return {"label": label, "color": color, "fillcolor": color, "fontcolor": font_c}

    def _input_output_params(
        self, node: ent.NodeModel, title: str
    ) -> t.Dict[str, t.Any]:
        rgb = colors.RGBColor(0, 0, 0)
        color = rgb.to_hex()
        font_c = self._text_color(rgb)

        label = self._multi_line(
            f'<B><FONT POINT-SIZE="{self._node_title_size}">{title}</FONT></B>',
            f"<I>{node.name}</I>",
        )
        return {"label": label, "color": color, "fillcolor": color, "fontcolor": font_c}

    def _input_params(self, node: ent.InputNodeModel) -> t.Dict[str, t.Any]:
        return self._input_output_params(node, "Input")

    def _output_params(self, node: ent.OutputNodeModel) -> t.Dict[str, t.Any]:
        return self._input_output_params(node, "Output")

    def _collapsed_params(self, node: ent.CollapsedNodeModel) -> t.Dict[str, t.Any]:
        rgb = self._color_picker.pick(*node.path[:-1])
        color = rgb.to_hex()
        font_c = self._text_color(rgb)

        joined_path = ".".join(node.path)
        joined_path = joined_path if len(joined_path) > 0 else "root"
        lines = [
            f'<B><FONT POINT-SIZE="{self._node_title_size}">{joined_path}</FONT></B>'
        ]

        if node.n_parameters is not None and node.n_parameters > 0:
            lines.append("")
            lines.append(
                f"<B>params</B>: {_human_format(node.n_parameters)} ({_human_format(node.perc_parameters*100)}%)"  # type: ignore
            )

        label = self._multi_line(*lines)

        return {"label": label, "color": color, "fillcolor": color, "fontcolor": font_c}

    def _node_params(self, node: ent.NodeModel) -> t.Dict[str, t.Any]:
        type_map = {
            "op": self._op_params,
            "input": self._input_params,
            "output": self._output_params,
            "collapsed": self._collapsed_params,
        }
        params = type_map[node.type_](node)
        return {**self._default_node_params, **params}

    def _subgraph_params(self, name: str, depth: int) -> t.Dict[str, t.Any]:
        # Bg color is a function of depth
        gray_level = int(255 * (1 / (depth / 10 + 1.1)))
        bgcolor = colors.RGBColor(gray_level, gray_level, gray_level).to_hex()

        # Label is the name of the subgraph
        label = f'<<B><FONT POINT-SIZE="{self._cluster_title_size}">{name.partition("cluster_")[2]}</FONT></B>>'

        params = {"label": label, "bgcolor": bgcolor, "labeljust": "l"}
        return {**self._default_subgraph_params, **params}

    def _get_tgt_graph_by_path(
        self, parent: pgv.AGraph, path: t.Sequence[str], depth: int = 0
    ) -> pgv.AGraph:
        # If the path is empty, return the parent graph
        if len(path) <= 1:
            return parent

        # If the subgraph does not exist, create it and recurse
        sg_name = "cluster_" + path[0]
        if parent.get_subgraph(sg_name) is None:
            parent.add_subgraph(name=sg_name, **self._subgraph_params(sg_name, depth))
        target = parent.get_subgraph(sg_name)
        assert target is not None
        return self._get_tgt_graph_by_path(target, path[1:], depth=depth + 1)

    def _edge_params(self, spec: t.Optional[dataspec.DataSpec]) -> t.Dict[str, t.Any]:
        if spec is None:
            return self._default_edge_params

        visitor = HTMLSpecVisitor()
        spec.accept(visitor)

        params = {"label": visitor.html}
        return {**params, **self._default_edge_params}

    def _convert(self, nngraph: ent.NNGraph) -> pgv.AGraph:
        # Initialize a pygraphviz graph
        pgvgraph = pgv.AGraph(directed=True, strict=True)

        # Populate nodes
        for node in nngraph.nodes:
            model = nngraph[node]
            target_graph = self._get_tgt_graph_by_path(pgvgraph, model.path)
            target_graph.add_node(node, **self._node_params(model))

        # Populate edges
        for source, target in nngraph.edges:
            spec = nngraph.get_spec(source, target)
            pgvgraph.add_edge(source, target, **self._edge_params(spec))

        return pgvgraph

    def draw(self, nngraph: ent.NNGraph) -> None:
        converted = self._convert(nngraph)
        prog = "dot"
        converted.draw(self._path, prog=prog)



from __future__ import annotations

import functools
import typing as t
from uuid import uuid4

import networkx as nx
import torch
import torch.fx as fx
import torch.nn as nn
from torchvision.models import feature_extraction

from nnviz import dataspec as ds
from nnviz import entities as ent
from nnviz import inspection as insp


class ExtendedFxGraph(fx.graph.Graph):
    """Extended version of `torch.fx.Graph` that adds some useful properties. Like:
    - `qualnames`: Mapping from node to qualified name.
    - `callables`: Mapping from node to its respecitve script/function/nn.Module.
    - `edges`: List of edges in the graph.
    """

    def __init__(
        self,
        wrapped: fx.graph.Graph,
        qualnames: t.Mapping[fx.node.Node, str],
        callables: t.Dict[fx.node.Node, nn.Module | t.Callable],
        specs: t.Optional[t.Mapping[fx.node.Node, ds.DataSpec]] = None,
    ) -> None:
        """Constructor.

        Args:
            wrapped (fx.graph.Graph): The wrapped graph.
            qualnames (t.Mapping[fx.node.Node, str]): A mapping from node to its qualified name.
            callables (t.Dict[fx.node.Node, nn.Module  |  t.Callable]): A mapping from node to its
                respective script/function/nn.Module.
            specs (t.Optional[t.Mapping[fx.node.Node, ds.DataSpec]], optional):
                A mapping from node to its output specs. Defaults to None.
        """
        super().__init__(
            wrapped.owning_module, wrapped._tracer_cls, wrapped._tracer_extras
        )
        self._wrapped = wrapped
        self._qualnames = qualnames
        self._callables = callables
        self._specs = specs or {}

    @property
    def nodes(self):
        """Returns the nodes in the graph."""
        return self._wrapped.nodes

    @property
    def total_parameters(self) -> int:
        """Returns the total number of parameters in the graph."""
        acc = 0
        for callable in self.callables.values():
            if isinstance(callable, nn.Module):
                acc += sum(p.numel() for p in callable.parameters())
        return acc

    def _recurse_args(self, args: t.Iterable[t.Any]) -> t.List[fx.node.Node]:
        deps = []
        for arg in args:
            if isinstance(arg, fx.node.Node):
                deps.append(arg)
                continue

            if isinstance(arg, t.Sequence) and not isinstance(arg, str):
                deps += self._recurse_args(arg)
            elif isinstance(arg, t.Mapping):  # pragma: no cover
                deps += self._recurse_args(arg.values())

        return deps

    def _get_dependencies(self, node: fx.node.Node) -> t.List[fx.node.Node]:
        return self._recurse_args([*node.args] + [*node.kwargs.values()])

    @functools.cached_property
    def edges(self) -> t.Sequence[t.Tuple[fx.node.Node, fx.node.Node]]:
        """Returns the edges in the graph as a list of tuples (arg, node)."""
        edges = []
        for node in self.nodes:
            for arg in self._get_dependencies(node):
                edges.append((arg, node))
        return edges

    @property
    def qualnames(self) -> t.Mapping[fx.node.Node, str]:
        """Returns a mapping from node to its qualified name."""
        return self._qualnames

    @property
    def callables(self) -> t.Dict[fx.node.Node, nn.Module | t.Callable]:
        """Returns a mapping from node to its respective script/function/nn.Module."""
        return self._callables

    @property
    def specs(self) -> t.Mapping[fx.node.Node, ds.DataSpec]:
        """Returns a mapping from node to its output data specs."""
        return self._specs


class ExtendedNodePathTracer(feature_extraction.NodePathTracer):
    """Extended version of `torchvision.models.feature_extraction.NodePathTracer` that generates
    an `ExtendedFxGraph` instead of a `torch.fx.Graph`.
    """

    def __init__(self, *args, inputs: t.Optional[t.Mapping[str, t.Any]] = None) -> None:
        super().__init__(*args)
        self._qualname_to_callable: t.Dict[str, t.Callable] = {}
        self._inputs = inputs

    def trace(
        self,
        root: nn.Module | t.Callable[..., t.Any],
        concrete_args: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> ExtendedFxGraph:
        # Reset the qualname to callable mapping
        self._qualname_to_callable = {}

        # Trace the graph
        wrapped = super().trace(root, concrete_args)
        gm = fx.graph_module.GraphModule(root, wrapped)  # type: ignore

        # Something is very wrong with the original torch.fx tracer. I am fixing it here:
        for k, v in self.node_to_qualname.items():
            k: fx.node.Node
            v: str

            # Ignore if the node is already in the mapping
            if v in self._qualname_to_callable:
                continue

            # Ignore if the node is either a get_attr or a placeholder
            if k.op in ["placeholder", "get_attr"]:
                continue

            # If the target is a string (as it should be) but no callable is found for it,
            # Then it probably means that the target is a reference to another node.
            # Example: when a layer is used multiple times in a model, every additional
            # time it is used, a new node is created that references the original node. These
            # nodes differ by their name (a _1, _2, etc... is appended to the name) but they
            # point to the same target.
            # I am tired of this fuck fiesta so I am just going to kill it with fire.
            if (
                isinstance(k.target, str)
                and v not in self._qualname_to_callable
                and k.target in self._qualname_to_callable
            ):
                self._qualname_to_callable[v] = self._qualname_to_callable[k.target]

            # If the target is a callable, then we can just add it to the mapping. This
            # is the case when the node is a builtin function (e.g. torch.add)
            # How does it even work? I don't know. I don't care. The less I know, the better.
            elif callable(k.target):
                self._qualname_to_callable[v] = k.target

        # Convert the callables (qualname -> callable) to a mapping (node -> callables)
        callables = {}
        for node in wrapped.nodes:
            qualname = self.node_to_qualname.get(node, "__INVALID__")
            callables[node] = self._qualname_to_callable.setdefault(
                qualname, node.target
            )

        specs = self._build_specs(gm)

        # Create the extended graph
        graph = ExtendedFxGraph(
            wrapped, qualnames=self.node_to_qualname, callables=callables, specs=specs
        )
        return graph

    def _build_specs(
        self, graph_module: fx.graph_module.GraphModule
    ) -> t.Optional[t.Mapping[fx.node.Node, ds.DataSpec]]:
        # Abort if no inputs are provided
        if self._inputs is None:
            return None

        # Spec collection
        specs: t.Dict[fx.node.Node, ds.DataSpec] = {}

        # Special cases
        special_cases = {
            "placeholder": lambda x: ds.DataSpec.build(self._inputs[x.name]),  # type: ignore
            "get_attr": lambda x: ds.DataSpec.build(
                graph_module.get_parameter(x.target)
            ),
        }

        # Collect all callables and their respective nodes
        node_clb_pairs: t.List[t.Tuple[fx.node.Node, t.Any]] = []
        for node, qualname in self.node_to_qualname.items():

            # Handle special cases first
            if node.op in special_cases:
                specs[node] = special_cases[node.op](node)

            # Register the node as a callable
            node_clb_pairs.append((node, self._qualname_to_callable[qualname]))

        # For the same reason as in `trace` we need to address the case where a nn.Module
        # is used multiple times in the graph. In this case, we need to register the specs
        # for all nodes that point to the same nn.Module.
        # Trust me, I don't like this either.
        def register_specs(the_self: nn.Module, input: t.Any, output: t.Any):
            tgt_nodes = [x for x, y in node_clb_pairs if y is the_self]
            specs.update({node: ds.DataSpec.build(output) for node in tgt_nodes})

        # It must be done for pure functions as well...
        def build_fn_wrapper(node: fx.node.Node, clb: t.Any):
            def wrapper(*args, **kwargs):
                output = clb(*args, **kwargs)
                specs[node] = ds.DataSpec.build(output)
                return output

            return wrapper

        # ... and for methods. Goddammit.
        def build_method_wrapper(node: fx.node.Node, target: str):
            def wrapper(*args, **kwargs):
                method = getattr(args[0], target)
                output = method(*args[1:], **kwargs)
                specs[node] = ds.DataSpec.build(output)
                return output

            return wrapper

        # Register the hooks
        for node, clb in node_clb_pairs:
            if isinstance(clb, nn.Module):
                clb.register_forward_hook(register_specs)
            elif callable(clb):
                node.target = build_fn_wrapper(node, clb)
            elif node.op == "call_method":
                node.op = "call_function"
                node.target = build_method_wrapper(node, node.target)  # type: ignore

        # This is the reason why people drink. To forget the horrors concealed in the depths of torch.fx.
        # The open source community is not going to miss me. And I don't blame them.
        # I need a therapist.
        graph_module = fx.graph_module.GraphModule(graph_module, graph_module.graph)

        # Run the model
        with torch.no_grad():
            graph_module(**self._inputs)

        return specs

    def call_module(self, m: nn.Module, forward: t.Callable, args, kwargs):
        old_qualname = self.current_module_qualname
        try:
            module_qualname = self.path_of_module(m)
            self._qualname_to_callable[module_qualname] = m
            self.current_module_qualname = module_qualname
            if not self.is_leaf_module(m, module_qualname):
                return forward(*args, **kwargs)
            return self.create_proxy("call_module", module_qualname, args, kwargs)
        finally:
            self.current_module_qualname = old_qualname


class TorchFxInspector(insp.NNInspector):
    """NNInspector implementation with torch.fx."""

    def inspect(
        self, model: nn.Module, inputs: t.Optional[t.Mapping[str, t.Any]] = None
    ) -> ent.NNGraph:
        tracer = ExtendedNodePathTracer(inputs=inputs)
        fxgraph = tracer.trace(model)
        return self._to_nngraph(fxgraph)

    def _convert_node(
        self, fxgraph: ExtendedFxGraph, node: fx.node.Node
    ) -> ent.NodeModel:
        node_map = {
            "call_module": self._op_node,
            "call_function": self._op_node,
            "call_method": self._op_node,
            "placeholder": self._input_node,
            "output": self._output_node,
            "get_attr": self._op_node,  # TODO: handle get_attr
        }

        return node_map[node.op](fxgraph, node)

    def _input_node(
        self, fxgraph: ExtendedFxGraph, node: fx.node.Node
    ) -> ent.NodeModel:
        node_name = fxgraph.qualnames.get(node, node.name)
        node_path = node_name.split(".")
        return ent.InputNodeModel(name=node_name, path=node_path)

    def _output_node(
        self, fxgraph: ExtendedFxGraph, node: fx.node.Node
    ) -> ent.NodeModel:
        node_name = fxgraph.qualnames.get(node, node.name)
        node_path = node_name.split(".")
        return ent.OutputNodeModel(name=node_name, path=node_path)

    def _extract_args(self, node: fx.node.Node) -> t.List[t.Any]:
        # Extract the constant arguments
        const_args = []
        for arg in node.args:
            if isinstance(arg, fx.node.Node):
                continue
            const_args.append(arg)

        return const_args

    def _extract_kwargs(self, node: fx.node.Node) -> t.Dict[str, t.Any]:
        # Extract the constant arguments
        const_kwargs = {}
        for k, v in node.kwargs.items():
            if isinstance(v, fx.node.Node):
                continue
            const_kwargs[k] = v

        return const_kwargs

    def _op_node(self, fxgraph: ExtendedFxGraph, node: fx.node.Node) -> ent.NodeModel:
        node_name = fxgraph.qualnames.get(node, node.name)
        callable_ = fxgraph.callables.get(node, node.target)
        node_path = node_name.split(".")
        n_params, perc_params = None, None
        if isinstance(callable_, nn.Module):
            op = callable_.__class__.__name__
            full_op = ".".join([str(callable_.__module__), op])
            n_params = sum(p.numel() for p in callable_.parameters())
            if fxgraph.total_parameters > 0:  # pragma: no branch
                perc_params = n_params / fxgraph.total_parameters
        elif isinstance(callable_, t.Callable):
            op = callable_.__name__
            full_op = ".".join([str(callable_.__module__), op])
        else:
            full_op = str(callable_)
            op = str(callable_)

        # Create the node model
        return ent.OpNodeModel(
            name=node_name,
            path=node_path,
            op=op,
            full_op=full_op,
            const_args=self._extract_args(node),
            const_kwargs=self._extract_kwargs(node),
            n_parameters=n_params,
            perc_parameters=perc_params,
        )

    def _to_nngraph(self, fxgraph: ExtendedFxGraph) -> ent.NNGraph:
        # Initialize a networkx graph
        nxgraph = nx.DiGraph()
        nngraph = ent.NNGraph(nxgraph)

        # Populate graph
        for src, tgt in fxgraph.edges:
            src_name = src.name if isinstance(src, fx.node.Node) else str(uuid4())
            tgt_name = tgt.name

            # Create the edge
            spec = fxgraph.specs.get(src, None)
            nngraph.add_edge(src_name, tgt_name, spec=spec)

            # Convert to node model both source and target
            src_model = self._convert_node(fxgraph, src)
            tgt_model = self._convert_node(fxgraph, tgt)

            # Update the graph
            nngraph[src_name] = src_model
            nngraph[tgt_name] = tgt_model

        return nngraph



from __future__ import annotations

import typing as t

import networkx as nx
import pydantic as pyd
from nnviz import dataspec


class NodeModel(pyd.BaseModel):
    """Pydantic model for a node in the graph. Contains some information about a layer of a neural network."""

    type_: t.Literal[""] = ""

    name: str = pyd.Field(..., description="Qualified name of the node.")
    path: t.Sequence[str] = pyd.Field(
        default_factory=list,
        description="Path of the node. E.g. ['features', '0', 'conv1']",
    )

    def depth(self) -> int:
        """Returns the depth of the node in the graph."""
        return len(self.path)


class OpNodeModel(NodeModel):
    """`NodeModel` specialized for an operation node."""

    type_: t.Literal["op"] = "op"

    op: str = pyd.Field("", description="Name of the operation. E.g. 'Conv2d'.")
    full_op: str = pyd.Field(
        "",
        description="Full name of the operation. E.g. 'torch.nn.modules.conv.Conv2d'.",
    )
    const_args: t.Sequence[t.Any] = pyd.Field(
        default_factory=list,
        description="Constant positional arguments of the operation.",
    )
    const_kwargs: t.Mapping[str, t.Any] = pyd.Field(
        default_factory=dict, description="Constant keyword arguments of the operation."
    )
    n_parameters: t.Optional[int] = pyd.Field(None, description="Number of parameters.")
    perc_parameters: t.Optional[float] = pyd.Field(
        None, description="Percentage of parameters."
    )


class CollapsedNodeModel(NodeModel):
    """`NodeModel` specialized for a collapsed node."""

    type_: t.Literal["collapsed"] = "collapsed"

    n_parameters: t.Optional[int] = pyd.Field(None, description="Number of parameters.")
    perc_parameters: t.Optional[float] = pyd.Field(
        None, description="Percentage of parameters."
    )


class InputNodeModel(NodeModel):
    """`NodeModel` specialized for an input node."""

    type_: t.Literal["input"] = "input"


class OutputNodeModel(NodeModel):
    """`NodeModel` specialized for an output node."""

    type_: t.Literal["output"] = "output"


class NNGraph:
    """Graph representation of a neural network."""

    MODEL_KEY = "model"
    """Key used to store the `NodeModel` in the graph nodes."""
    SPEC_KEY = "spec"
    """Key used to store the `DataSpec` in the graph edges."""

    @classmethod
    def empty(cls) -> NNGraph:
        """Create an empty graph."""
        return cls(nx.DiGraph())

    def __init__(self, graph: nx.DiGraph) -> None:
        """Constructor.

        Args:
            graph (nx.DiGraph): NetworkX graph wrapped by this class.
        """
        self._graph = graph

    def __getitem__(self, name: str) -> NodeModel:
        return self._graph.nodes[name][self.MODEL_KEY]

    def __setitem__(self, name: str, model: NodeModel) -> None:
        if name in self._graph.nodes:
            self._graph.nodes[name][self.MODEL_KEY] = model
        else:
            self._graph.add_node(name, model=model)

    @property
    def nodes(self) -> t.Iterable[str]:
        """Returns an iterator over the nodes in the graph."""
        yield from self._graph.nodes

    @property
    def edges(self) -> t.Iterable[t.Tuple[str, str]]:
        """Returns an iterator over the edges in the graph."""
        yield from self._graph.edges

    def add_edge(
        self, source: str, target: str, spec: t.Optional[dataspec.DataSpec] = None
    ) -> None:
        """Add an edge to the graph.

        Args:
            source (str): The source node.
            target (str): The target node.
            spec (t.Optional[DataSpec], optional): The data spec associated with the edge. Defaults to None.
        """
        if (source, target) in self._graph.edges:
            self._graph.edges[source, target][self.SPEC_KEY] = spec
        self._graph.add_edge(source, target, spec=spec)

    def get_spec(self, source: str, target: str) -> t.Optional[dataspec.DataSpec]:
        """Get the data spec associated with an edge.

        Args:
            source (str): The source node.
            target (str): The target node.

        Returns:
            t.Optional[DataSpec]: The data spec associated with the edge.
        """
        return self._graph.edges[source, target].get(self.SPEC_KEY, None)

    # TODO: refactor this method
    def collapse(self, depth: int) -> NNGraph:
        """Collapse the graph by grouping nodes at the same level.

        Args:
            depth (int): The depth at which the nodes should be collapsed. If < 0, no collapse is performed.

        Returns:
            NNGraph: The collapsed graph.
        """
        if depth < 0:
            return self

        # Create a mapping (path -> node) for each node in the graph
        path_to_node = {}
        for node in self.nodes:
            node_model = self[node]
            path = tuple(node_model.path[:depth])
            path_to_node[path] = node

        # Create a mapping (node -> collapsed node)
        node_to_collapsed: t.Dict[str, t.Tuple[str, NodeModel]] = {}
        collapsed_nodes: t.Dict[str, CollapsedNodeModel] = {}
        for node in self.nodes:
            node_model = self[node]
            if node_model.depth() <= depth:
                node_to_collapsed[node] = node, node_model
            else:
                path = tuple(node_model.path[:depth])
                collapsed_node = path_to_node[path]
                if collapsed_node not in collapsed_nodes:
                    collapsed_nodes[collapsed_node] = CollapsedNodeModel(
                        name=collapsed_node,
                        path=path,
                        n_parameters=0,
                        perc_parameters=0,
                    )
                collapsed_model = collapsed_nodes[collapsed_node]

                if (
                    isinstance(node_model, OpNodeModel)
                    and node_model.n_parameters is not None
                ):
                    collapsed_model.n_parameters += node_model.n_parameters  # type: ignore
                    collapsed_model.perc_parameters += node_model.perc_parameters  # type: ignore
                node_to_collapsed[node] = collapsed_node, collapsed_model

        # Create a new graph
        collapsed = self.empty()

        # Populate the new graph
        for source, target in self.edges:
            collapsed_src, model_src = node_to_collapsed[source]
            collapsed_tgt, model_tgt = node_to_collapsed[target]
            collapsed[collapsed_src] = model_src
            collapsed[collapsed_tgt] = model_tgt

            collapsed_spec = None
            if (collapsed_src, collapsed_tgt) in self.edges:
                collapsed_spec = self.get_spec(collapsed_src, collapsed_tgt)
            elif collapsed_tgt != collapsed_src:
                # Find all the edges that start from the collapsed source
                collapsed_src_edges = [
                    (src, tgt)
                    for src, tgt in self.edges
                    if src == collapsed_src and tgt != collapsed_src
                ]

                # Gather specs from all the edges that start from the collapsed source
                collapsed_src_specs = [
                    self.get_spec(src, tgt) for src, tgt in collapsed_src_edges
                ]

                # Find the first spec that is not None
                collapsed_spec = next(
                    (spec for spec in collapsed_src_specs if spec is not None), None
                )

            # Add the edge only if it is not a self-loop
            if collapsed_src != collapsed_tgt:
                collapsed.add_edge(collapsed_src, collapsed_tgt, spec=collapsed_spec)

        return collapsed



from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod

color_t = t.Tuple[int, int, int]


class RGBColor:
    """A class that represents a color in RGB format."""

    @classmethod
    def from_hex(cls, hex_color: str) -> RGBColor:
        """Creates a new RGBColor from a hex color in the format #RRGGBB.

        Args:
            hex_color (str): The hex color to convert.

        Returns:
            RGBColor: The RGBColor representation of the hex color.
        """
        r, g, b = tuple(int(hex_color.lstrip("#")[i : i + 2], 16) for i in (0, 2, 4))
        return cls(r, g, b)

    def __init__(self, r: int, g: int, b: int) -> None:
        """Creates a new RGBColor.

        Args:
            r (int): The red component of the color.
            g (int): The green component of the color.
            b (int): The blue component of the color.
        """
        self._color = (r, g, b)

    def is_bright(self) -> bool:
        """Returns whether the color is bright or not.

        Returns:
            bool: True if the color is bright, False otherwise.
        """
        r, g, b = self._color
        return (r * 0.299 + g * 0.587 + b * 0.114) > 186

    def to_hex(self) -> str:
        """Converts the color to a hex color in the format #RRGGBB.

        Returns:
            str: The hex color.
        """
        return "#%02x%02x%02x" % self._color


class ColorPicker(ABC):
    """An abstract class that represents a color picker."""

    @abstractmethod
    def pick(self, *args: t.Hashable) -> RGBColor:
        """Pick a color based on the given arguments.

        Returns:
            RGBColor: The color picked.
        """
        pass



import typing as t
from pathlib import Path

import typer

app = typer.Typer(name="nnviz", pretty_exceptions_enable=False)

model_help = """
The model to visualize. Can either be: \n
- The name of a model in `torchvision.models`. \n
- A string in the form `PATH_TO_FILE:NAME`, where `PATH_TO_FILE` is the path to a python
file containing a model, and `NAME` is the name of the model in that file. Valid names
include objects of type `torch.nn.Module` and functions that return an object of type
`torch.nn.Module` with no arguments. Class constructors are considered like functions.
"""
out_help = "The output file path. If not provided, it will save a pdf file named after the model in the current directory."
depth_help = "The maximum depth of the graph. No limit if < 0."
show_help = "Also show the graph after drawing using the default pdf application."
input_help = """The input to feed to the model. If specified, nnviz will also add synthetic
representation of the data passing through the model. Can either be: \n
- "default" -> float32 BCHW tensor of shape (1, 3, 224, 224) (commonly used) \n
- "image<side>" (e.g. image224, image256, ...) -> float32 BCHH tensor \n
- "image<height>x<width>" (e.g. image224x224, image256x512, ...) -> float32 BCHW tensor \n
- "tensor<s0>x<s1>x<s2>x..." (e.g. tensor1x3x224x224, tensor1x3x256x512, ...) ->
float32 generic tensors \n
- "<key1>:<value1>;<key2>:<value2>;... (e.g. x:tensor1x3x224x224;y:tensor1x3x256x512,
...) -> dictionary of tensors \n
- A plain python string that evaluates to a dictionary of tensors (e.g. "{'x': torch.rand(1, 3, 224, 224)}")
"""


@app.command(name="quick")
def quick(
    model: str = typer.Argument(..., help=model_help),
    output_path: t.Optional[Path] = typer.Option(None, "-o", "--out", help=out_help),
    depth: int = typer.Option(1, "-d", "--depth", help=depth_help),
    show: bool = typer.Option(False, "-s", "--show", help=show_help),
    input: str = typer.Option(None, "-i", "--input", help=input_help),
) -> None:
    """Quickly visualize a model."""
    from nnviz import drawing, inspection

    def _show(output_path: Path) -> None:
        import os
        import subprocess
        import sys

        # Windows
        if os.name == "nt":
            os.startfile(output_path, "open")

        # Mac
        elif sys.platform == "darwin":
            subprocess.Popen(["open", output_path])

        # Linux
        elif sys.platform.startswith("linux"):
            subprocess.Popen(["xdg-open", output_path])

        else:
            typer.echo(f"Could not open {output_path} automatically. I'm sorry.")

    if output_path is None:
        _, _, model_name = model.rpartition(":")
        output_path = Path(f"{model_name}.pdf")

    # Load model
    try:
        nn_model = inspection.load_from_string(model)
    except Exception as e:
        raise typer.BadParameter(f"Could not load model {model}") from e

    # TODO: The inspector and drawer should be configurable, not hardcoded
    inspector = inspection.TorchFxInspector()
    drawer = drawing.GraphvizDrawer(output_path)

    # Parse input
    parsed_input = inspection.parse_input_str(input)

    # Inspect
    graph = inspector.inspect(nn_model, inputs=parsed_input)

    # Collapse by depth
    graph = graph.collapse(depth)

    # Draw
    drawer.draw(graph)

    # Show
    if show:
        _show(output_path)


if __name__ == "__main__":
    app()



from __future__ import annotations
from abc import ABC, abstractmethod

import typing as t

import pydantic as pyd
import torch


class DataSpecVisitor(ABC):
    """Visitor for the `DataSpec` class."""

    @abstractmethod
    def visit_tensor_spec(self, spec: TensorSpec) -> None:
        """Visits a `TensorSpec`."""
        pass

    @abstractmethod
    def visit_builtin_spec(self, spec: BuiltInSpec) -> None:
        """Visits a `BuiltInSpec`."""
        pass

    @abstractmethod
    def visit_list_spec(self, spec: ListSpec) -> None:
        """Visits a `ListSpec`."""
        pass

    @abstractmethod
    def visit_map_spec(self, spec: MapSpec) -> None:
        """Visits a `MapSpec`."""
        pass

    @abstractmethod
    def visit_unknown_spec(self, spec: UnknownSpec) -> None:
        """Visits a `UnknownSpec`."""
        pass


class DataSpec(pyd.BaseModel, ABC):
    """Specification of the data that is passed through the graph."""

    class Config:
        frozen = True

    spec_type: t.Literal[""] = ""

    @abstractmethod
    def accept(self, visitor: DataSpecVisitor) -> None:
        """Accepts an incoming visitor."""
        pass

    @classmethod
    def build(cls, data: t.Any) -> DataSpec:
        if isinstance(data, torch.Tensor):
            return TensorSpec(shape=data.shape, dtype=str(data.dtype))
        elif isinstance(data, (list, tuple)):
            return ListSpec(elements=[cls.build(x) for x in data])
        elif isinstance(data, dict):
            return MapSpec(elements={k: cls.build(v) for k, v in data.items()})
        elif isinstance(data, (int, float, str, bool, type(None), type(...), bytes)):
            return BuiltInSpec(name=data.__class__.__name__)
        else:
            return UnknownSpec()


class TensorSpec(DataSpec):
    """Specification of the data that is passed through the graph."""

    spec_type: t.Literal["tensor"] = "tensor"
    shape: t.Sequence[int] = pyd.Field(
        default_factory=list, description="Shape of the tensor."
    )
    dtype: str = pyd.Field("", description="Data type of the tensor.")

    def accept(self, visitor: t.Any) -> None:
        visitor.visit_tensor_spec(self)


class BuiltInSpec(DataSpec):
    """Specification of the data that is passed through the graph."""

    spec_type: t.Literal["builtin"] = "builtin"
    name: str = pyd.Field("", description="Name of the builtin type.")

    def accept(self, visitor: t.Any) -> None:
        visitor.visit_builtin_spec(self)


class UnknownSpec(DataSpec):
    """Specification of the data that is passed through the graph."""

    spec_type: t.Literal["unknown"] = "unknown"

    def accept(self, visitor: t.Any) -> None:
        visitor.visit_unknown_spec(self)


class ListSpec(DataSpec):
    """Specification of the data that is passed through the graph."""

    spec_type: t.Literal["list"] = "list"

    elements: t.Sequence[DataSpec] = pyd.Field(
        default_factory=list, description="List of elements in the list."
    )

    @pyd.validator("elements")
    def validate_elements(cls, v):
        return [pyd.parse_obj_as(t_any_spec, v) for v in v]

    def accept(self, visitor: t.Any) -> None:
        visitor.visit_list_spec(self)
        for element in self.elements:
            element.accept(visitor)


class MapSpec(DataSpec):
    """Specification of the data that is passed through the graph."""

    spec_type: t.Literal["map"] = "map"

    elements: t.Mapping[str, DataSpec] = pyd.Field(
        default_factory=dict, description="Mapping of keys to elements in the map."
    )

    @pyd.validator("elements")
    def validate_elements(cls, v):
        return {k: pyd.parse_obj_as(t_any_spec, v) for k, v in v.items()}

    def accept(self, visitor: t.Any) -> None:
        visitor.visit_map_spec(self)
        for element in self.elements.values():
            element.accept(visitor)


t_any_spec = t.Union[TensorSpec, BuiltInSpec, ListSpec, MapSpec]



import hashlib
import typing as t

from nnviz import colors


class HashColorPicker(colors.ColorPicker):
    """A color picker that uses a hash function to pick a color."""

    def pick(self, *args: t.Hashable) -> colors.RGBColor:
        hash_ = hashlib.sha256(str(args).encode("utf-8")).hexdigest()
        index = int(hash_, 16) % (2**24)
        r = (index >> 16) & 0xFF
        g = (index >> 8) & 0xFF
        b = index & 0xFF
        return colors.RGBColor(r, g, b)



from abc import ABC, abstractmethod

import torch.nn as nn

from nnviz import entities as ent


class NNInspector(ABC):
    """An abstract class for inspecting a neural network model. All nnviz inspectors
    should inherit from this class.
    """

    @abstractmethod
    def inspect(self, model: nn.Module) -> ent.NNGraph:
        """Inspect a neural network model and return a NNGraph object.

        Args:
            model (nn.Module): The neural network model to inspect.

        Returns:
            NNGraph: The graph representation of the neural network model.
        """
        pass



from abc import ABC, abstractmethod

from nnviz import entities as ent


class GraphDrawer(ABC):
    """An abstract class for drawing a neural network graph. All nnviz graph drawers
    should inherit from this class.
    """

    @abstractmethod
    def draw(self, nngraph: ent.NNGraph) -> None:
        """Draw a neural network graph. This method returns nothing, to avoid making decisions
        about where to save/display/return the graph. The implementation of this interface
        can add custom logic to save/display/return the graph separately.

        Args:
            nngraph (NNGraph): The neural network graph to draw.
        """
        pass
